module SSG.Djot.Inline

import SSG.Parser.Core
import SSG.Parser.Util

import SSG.Djot.Common

import Control.Eff
import Derive.Prelude

-- For iutils unit tests
import System

%language ElabReflection

--**************
--* Data Types *
--**************

public export
data Inline : Type where
  HardLineBreak : Inline
  SoftLineBreak : Inline
  NonBreakingSpace : Inline
  Text : (c : Char) -> Inline

%runElab derive "Inline" [Show, Eq]

--**************
--*   Syntax   *
--**************

------------------------
-- Escaped Whitespace --
------------------------

hardLineBreak : PS Inline
hardLineBreak = do
  _ <- thisString "\\"
  _ <- spaces
  _ <- lineEnding
  pure $ HardLineBreak

nbsp : PS Inline
nbsp = do
  _ <- thisString "\\ "
  pure $ NonBreakingSpace

softLineBreak : PS Inline
softLineBreak = do
  _ <- lineEnding
  -- Check to see if the next line is empty, if it is, we are at the end of the inline
  -- content, go ahead and bail
  state <- save
  Nothing <- tryMaybe blankLine
    | Just _ => throw "End of inline"
  load state
  pure $ SoftLineBreak

---------------------
-- Emphasis/Strong --
---------------------

----------
-- Text --
----------

-- Process escape codes before processing as text
escapedText : PS Inline
escapedText = do
  c <- escapedChar
  pure $ Text c

-- Any non-line-ending character can be part of regular text
text : PS Inline
text = do
  c <- nonTerminal
  pure $ Text c

---------------------------
-- Overall Inline Parser --
---------------------------

inlineElementsNoNewlines : PS Inline
inlineElementsNoNewlines = oneOfE "" [
    nbsp
  , escapedText
  -- Text is last so that anything can superseed it
  , text
  ]

inlineElement : PS Inline
inlineElement = oneOfE "" [
    hardLineBreak
  , softLineBreak
  , inlineElementsNoNewlines
  ]

export
inline : PS (List1 Inline)
inline = atLeastOne "Expected Inline Content" inlineElement

--**************
--* Unit Tests *
--**************

-------------------------------
-- Testing Utility Functions --
-------------------------------

export
inlineFromString : String -> List (Inline)
inlineFromString str with (asList str)
  inlineFromString "" | [] = []
  inlineFromString (strCons c str) | (c :: x) =
    Text c :: inlineFromString str | x

export
inlineFromString' : String -> List1 (Inline)
inlineFromString' str =
  case inlineFromString str of
    [] => assert_total $ idris_crash "Bad unit test fromString"
    (x :: xs) => x ::: xs

-----------
-- Tests --
-----------

-- @@test Plain Text Hello World
inlineTextSmoke : IO Bool
inlineTextSmoke =
  let input = "Hello World!" in
  golden input (map Text . unpack $ input) (map forget inline)

-- @@test Escaped Text
inlineEscapedSmoke : IO Bool
inlineEscapedSmoke =
  let input = "Hello\\n\\*World"
      ref = inlineFromString "Hello" ++ [Text '\n', Text '*'] ++ inlineFromString "World"
  in golden input ref (map forget inline)

-- @@test Hard Line Break
inlineHardBreakSmoke : IO Bool
inlineHardBreakSmoke =
  let input = "Hello\\\nWorld"
      ref = inlineFromString "Hello" ++ [HardLineBreak] ++ inlineFromString "World"
  in golden input ref (map forget inline)

-- @@test Soft Line Break
inlineSoftBreakSmoke : IO Bool
inlineSoftBreakSmoke =
  let input = "Hello\nWorld"
      ref = inlineFromString "Hello" ++ [SoftLineBreak] ++ inlineFromString "World"
  in golden input ref (map forget inline)

-- @@test Nonbreaking Space
inlineNbspSmoke : IO Bool
inlineNbspSmoke =
  let input = "Hello\\ World"
      ref = inlineFromString "Hello" ++ [NonBreakingSpace] ++ inlineFromString "World"
  in golden input ref (map forget inline)
